/************************************************************************
* EPP RTK C++
* Copyright (C) 2001 Global Name Registry
*
* This library is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published
* by the Free Software Foundation; either version 2.1 of the License, or
* (at your option) any later version.
*
* This library is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Contact information:  epprtk@gnr.com
*
*                       EPP RTK
*                       GNR Ltd.
*                       125 High Holborn
*                       London WC1V 6QA
*                       United Kingdom
************************************************************************/

#include "dom_tools.h"
#include "dom_output.h"
#include <string>

using namespace std;

/// Do XML entity encoding when printing... 
int domtools::basic_output::doEncode = domtools::basic_output::ENCODE_PARTIAL;

static const string amp = "&amp;";

static const string & encodeString(const string & text, string & encoded)
{
   if (domtools::basic_output::doEncode == domtools::basic_output::ENCODE_NONE) return text;

   bool partial = domtools::basic_output::doEncode == domtools::basic_output::ENCODE_PARTIAL;

   encoded = "";
   string::const_iterator i = text.begin();
   while (i != text.end()) {
      switch (*i) {
       case '&':
	 if (partial) {
	    // Could be part of a legal entity, so check:
	    string::const_iterator j = i;
	    j++;
	    // If '&' is at the end of the string, it is not a legal entity
	    if (j == text.end()) {
	       encoded += amp; break;
	    }
	    
	    unsigned int k = 0;
	    
	    // Could be numeric entity. If it is not, then '&' must be escaped
	    if (*j == '#') {
	       j++;
	       if (j == text.end()) {
		  encoded += amp; break;
	       }
	       
	       if (*j == 'x') {
		  // hex
		  j++;
		  static const char * hexdigits = "0123456789abcdefABCDEF";
		  while (j != text.end() && *j != ';' && strchr(hexdigits,*j)) {
		     j++; k++;
		  }
		  if (j != text.end() && *j == ';' && k > 0) encoded += "&";
		  else encoded = amp;
		  break;
	       };
	       
	       if (!isdigit(*j)) {
		  encoded += amp; break;
	       };
	       
	       // decimal
	       while (j != text.end() && *j != ';' && isdigit(*j)) {
		  j++; k++;
	       }
	       if (j != text.end() && *j == ';' && k > 0) encoded += "&";
	       else encoded += amp;
	       break;
	    }
	    
	    // Check the standard entites, amp, apos, lt, gt, quot:
	    string entity;
	    while (j != text.end() && *j != ';' && k < 5) {
	       entity += *j;
	       j++; k++;
	    }

	    // Was there a semicolon at the end?
	    if (j == text.end()) {
	       encoded += amp;
	       break;
	    }
	    
	    if (entity == "amp") encoded += '&';
	    else if (entity == "gt") encoded += '&';
	    else if (entity == "lt") encoded += '&';
	    else if (entity == "quot") encoded += '&';
	    else if (entity == "apos") encoded += '&';
	    else encoded += amp;
	 } else encoded += amp;
	 break;
       case '<':
	 encoded += "&lt;";
	 break;
       case '>':
	 encoded += "&gt;";
	 break;
       case '"':
	 encoded += "&quot;";
	 break;
       case '\'':
	 encoded += "&apos;";
	 break;
       default:
	 encoded += *i;
      };
      i++;
   }
   return encoded;
}


void domtools::basic_output::putTag(const string & tag, const string & cdata)
{
   beginTag(tag);
   if (cdata != "") putCDATA(cdata);
   endTag(tag);
}


void domtools::xml_output::outputWhitespace(int num)
{
   string tmp(num,' ');
   outputText(tmp);
}

void domtools::xml_output::beginTag(const string & tag)
{
   if (debug) cerr << "xml_output::beginTag(" << tag << ")" << endl;
   
   if (inPI) DOMTOOLS_EXCEPTION("Should not call beginTag() while in a processing instruction");
   if (nest > 0 && !hasContent) outputXML(">");
   
   if (ws && !firsttag) {
      if (nest) outputLF();
      outputWhitespace(nest*nestws);
   }
   
   outputXML("<");

   outputText(tag);
   hasContent = false;
   hasChildren = false;
   firsttag = false;
   lastattrcol = 0;
   nest++;
}

void domtools::xml_output::beginPI(const string & tag)
{
   // NOTE: We are making the arrogant assumption that nobody will want to put
   // anything but attributes in a PI. 
   
   if (debug) cerr << "xml_output::beginPI(" << tag << ")" << endl;
   
   if (inPI) DOMTOOLS_EXCEPTION("Should not call beginPI() when already in a processing instruction");
   
   if (ws && !firsttag) {
      if (nest) outputLF();
      outputWhitespace(nest*nestws);
   }
   
   outputXML("<?");
   
   outputText(tag);
   hasContent = false;
   hasChildren = false;
   firsttag = false;
   lastattrcol = 0;
   nest++;
   
   inPI = true;
}
	
void domtools::xml_output::putAttribute(const string & key)
{
   if (debug) cerr << "xml_output::putAttribute(" << key << ")" << endl;
   
   if (nest < 1) DOMTOOLS_EXCEPTION("Can not put attributes outside tags");
   if (hasContent) DOMTOOLS_EXCEPTION("Can not put attributes to a tag after CDATA or text content has been added");
   
   if (!inPI && ws && (lastattrcol != 0)) {
      outputLF();
      outputWhitespace(lastattrcol);
   } else {
      outputXML(" ");
      lastattrcol = col;
   }
   
   outputText(key);
}

void domtools::xml_output::putAttribute(const string & key, const string & value)
{
   putAttribute(key);
   if (debug) cerr << "                           value=\"" << value << "\"" << endl;
   
   outputXML("=\"");
   string encoded;
   outputText(encodeString(value,encoded));
   outputXML("\"");
}


void domtools::xml_output::endPI(void)
{
   if (debug) cerr << "xml_output::endPI()" << endl;
   if (nest < 1) DOMTOOLS_EXCEPTION("Lacks beginPI()");
   
   nest--;
   inPI = false;
   hasContent = true;
   hasChildren = true;
   outputXML("?>");
   if (!nest) outputLF();
   return;
}


void domtools::xml_output::endTag(const string & tag)
{
   if (debug) cerr << "xml_output::endTag()" << endl;
   if (nest < 1) DOMTOOLS_EXCEPTION("Lacks beginTag()");
   if (!hasContent) outputXML("/>");
   else {
      if (ws && hasChildren) {
	 outputLF();
	 outputWhitespace((nest-1) * nestws);
      }
      
      outputXML("</");
      outputText(tag);
      outputXML(">");
   }
   
   nest--;
   if (!nest) outputLF();
   hasContent = true;
   hasChildren = true;
}




void domtools::xml_output::putCDATA(const string & text)
{
   if (debug) cerr << "xml_output::putCDATA(" << text << ")" << endl;
   if (!inPI && !hasContent) {
      if (nest != 0) outputXML(">");
      hasContent = true;
   }

   string encoded;
   outputText(encodeString(text,encoded));
}
