/************************************************************************
* EPP RTK C++
* Copyright (C) 2001 Global Name Registry
*
* This library is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published
* by the Free Software Foundation; either version 2.1 of the License, or
* (at your option) any later version.
*
* This library is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Contact information:  epprtk@gnr.com
*
*                       EPP RTK
*                       GNR Ltd.
*                       125 High Holborn
*                       London WC1V 6QA
*                       United Kingdom
************************************************************************/

#ifndef __DOM_WRAPPER_H
#define __DOM_WRAPPER_H


// \file Encapsulates XML parser specific functionality. Supports many DOM functions, plus additional utility functions.

#include "refcount.h"
#include "dom_tools.h"
#include "dom_output.h"

#ifdef HAVE_XERCESC_PARSERS_DOMPARSER_HPP
# undef HAVE_XERCES
# define HAVE_XERCES 1
#endif

#ifdef HAVE_XERCES
# include <xercesc/framework/MemBufInputSource.hpp>
#endif


namespace domtools {

   class Node;

   /// Used for attribute maps etc.
   class NamedNodeMap {
    public:
      virtual ~NamedNodeMap(void) { }
      virtual Node * item(int pos) = 0;
      virtual int getLength(void) = 0;
   };
   
   /** A partially DOM compliant interface with additional utility functions
    * 
    *  Methods present in the DOM specifications are marked as such. 
    *  Expect these classes to deviate somewhat from the DOM even for those
    *  methods. For instance, we use standard C++ strings. 
    * 
    *  When it comes to conventions for interpreting the DOM IDLs for C++,
    *  this class mostly follows Xerces.
    */
   
   class Node {
    public:
      static const int DOCUMENT_NODE = 0;
      static const int XML_DECL_NODE = 1;
      static const int PROCESSING_INSTRUCTION_NODE = 2;
      static const int ELEMENT_NODE = 3;
      static const int CDATA_SECTION_NODE = 4;
      static const int TEXT_NODE = 5;
      static const int ATTRIBUTE_NODE = 6;
      
      virtual ~Node(void) {}
      
      // --------------------------------------------------------------------
      // Methods implementing (parts of) DOM Level 1 CORE interface "Node":
      
      /** Get the type of this node
       *  DOM level 1 interface Node
       */
      virtual unsigned short getNodeType(void) = 0;

      /// Return the "local" part of the nodes name (namespace identifier excluded)
      virtual std::string getLocalName(void) = 0;

      /// DOM function getFirstChild
      virtual Node * getFirstChild(void) = 0;
      
      /// DOM function
      virtual Node * getNextSibling(void) = 0;
      
      /// DOM function
      virtual std::string getNodeName(void) = 0;
      
      /// DOM function
      virtual std::string getNodeValue(void) = 0;

      virtual NamedNodeMap * getAttributes(void) = 0;

      
      
      // --------------------------------------------------------------------
      // From the element interface
      
      /// Get a named attribute of the current node.
      virtual std::string getAttribute(const std::string &) = 0;

      
      // --------------------------------------------------------------------

      
      
      /// Determine whether a node actually reference a value or not.
      virtual bool isNull(void) = 0;
      
      /// Return the first child node of the current node that is a DOM element.
      virtual Node * getFirstChildElement(void) = 0;
      
      /// Get the data of the first child element, or an empty string if the element does not exist.
      virtual std::string getFirstChildElementData(void) = 0;

      /// Return the namespace URI
      virtual std::string getNamespaceURI(void) = 0;

      /// Return the namespace prefix
      virtual std::string getPrefix(void) = 0;

      /// Return the node data
      virtual std::string getNodeData(void) = 0;
      
      /// Return the first sibling of this node that is an element
      virtual Node * getNextSiblingElement(void) = 0;

      /// Get the tag named by the string argument, searching from the document root of the current node
      virtual Node * getTag(const std::string &) = 0;  
      
      /// Get the data of the node for the tag represented by the string argument, searching from the document root
      virtual std::string getTagData(const std::string & tag) = 0;
      
      /// Return the first child of the tag represented by the argument, searching from the document root
      virtual Node * getFirstChildOfTag(const std::string &) = 0;
      
      
      
      // XML Decl specific
      virtual std::string getVersion(void) = 0;
      virtual std::string getEncoding(void) = 0;
      virtual std::string getStandalone(void) = 0;
   };
   
   /// Implements generic functionality that does not have to be provided for every xml parser
   class GenericNode : public Node {
    public:
      virtual ~GenericNode() { }
      virtual std::string getFirstChildElementData(void);
      
      // These functions work on a tag referenced from the document root of the current node
      virtual std::string getTagData(const std::string & tag);
   };


   /// Use dom_ptr to simplify storage management. 
   typedef refcnt_ptr<domtools::Node> dom_ptr;
   typedef refcnt_ptr<domtools::NamedNodeMap> dom_namedmap_ptr;
   
   
   class Parser {
    public:
      virtual ~Parser() { }
      virtual Node * parseFile(const std::string & filename) = 0;
      virtual Node * parseMemory(const void * const buffer, int length) = 0;
      virtual Node * parseString(const std::string & str);
   };
   
#ifdef HAVE_XERCES
   // Xerces support

   class XercesNode;
   
   /// Implementation of domtools::NamedNodeMap for the Xerces XML parser
   class XercesNamedNodeMap : public NamedNodeMap {
    private:
      DOM_NamedNodeMap node;
    public:
      explicit XercesNamedNodeMap(const DOM_NamedNodeMap & n) : node(n)
	{
	}

      XercesNamedNodeMap(void)
	{
	}
      
      virtual ~XercesNamedNodeMap()
	{
	}
      
      virtual Node * item(int pos);
      virtual int getLength(void);
      
   };
   
   /// Implementation of domtools::Node for the Xerces XML parser
   class XercesNode : public GenericNode {      
    private:
      DOM_Node node;
    public:
      explicit XercesNode(const DOM_Node & n) : node(n)
	{
	}

      XercesNode(void)
	{
	}
      
      virtual bool isNull(void);
      virtual std::string getLocalName(void);
      
      virtual ~XercesNode(void)
	{
	}


      virtual std::string getNodeData(void)
	{
	   if (isNull()) return "";
	   std::string tmp;
	   domtools::getNodeData(node,tmp);
	   return tmp;
	}
      
      virtual Node * getFirstChildElement(void) 
	{
	   if (isNull()) return new XercesNode();
	   return new XercesNode(domtools::getFirstChildElement(node));
	}
      
      virtual Node * getNextSiblingElement(void) 
	{
	   if (isNull()) return new XercesNode();
	   return new XercesNode(domtools::getNextSiblingElement(node));
	}

      virtual Node * getFirstChildOfTag(const std::string & tag) 
	{
	   if (isNull()) return new XercesNode();
	   return new XercesNode(domtools::getFirstChildOfTag(node,tag));
	}

      virtual Node * getTag(const std::string & tag)
	{
	   if (isNull()) return new XercesNode();
	   return new XercesNode(domtools::getSingleTag(node,tag));
	}

      virtual std::string getAttribute(const std::string & attr)
	{
	   if (isNull()) return "";
	   if (node.getNodeType() != DOM_Node::ELEMENT_NODE) return "";
	   DOM_Element e = (DOM_Element &)node;
	   return toString(e.getAttribute(attr.c_str()));
	}

      virtual Node * getFirstChild(void)
	{
	   if (isNull()) return new XercesNode();
	   return new XercesNode(node.getFirstChild());
	}
      
      virtual std::string getNamespaceURI(void)
	{
	   if (isNull()) return "";
	   return toString(node.getNamespaceURI());
	}

      virtual std::string getPrefix(void)
	{
	   if (isNull()) return "";
	   return toString(node.getPrefix());
	}

      virtual Node * getNextSibling(void)
	{
	   if (isNull()) return new XercesNode();
	   return new XercesNode(node.getNextSibling());
	}

      virtual std::string getNodeValue(void)
	{
	   if (isNull()) return "";
	   return toString(node.getNodeValue());
	}

      virtual std::string getNodeName(void)
	{
	   if (isNull()) return "";
	   return toString(node.getNodeName());
	}

      virtual NamedNodeMap * getAttributes(void)
	{
	   if (isNull()) return 0;
	   return new XercesNamedNodeMap(node.getAttributes());
	}
      
      virtual unsigned short getNodeType(void);
      
      
      virtual std::string getVersion(void)
	{
	   if (node.isNull() || (node.getNodeType() != DOM_Node::XML_DECL_NODE)) return "";
	   return toString(((DOM_XMLDecl &)node).getVersion());
	}
      
      virtual std::string getEncoding(void)
	{
	   if (node.isNull() || (node.getNodeType() != DOM_Node::XML_DECL_NODE)) return "";
	   return toString(((DOM_XMLDecl &)node).getEncoding());
	}
      
      virtual std::string getStandalone(void)
	{
	   if (node.isNull() || (node.getNodeType() != DOM_Node::XML_DECL_NODE)) return "";
	   return toString(((DOM_XMLDecl &)node).getStandalone());
	}
   };

   /// Implementation of domtools::Parser for the Xerces XML parser
   class XercesParser : public Parser {
    public:
      
      XercesParser()
	{
	}
      
      ~XercesParser()
	{
	}
      
      virtual Node * parseFile(const std::string & filename);
      virtual Node * parseMemory(const void * const buffer, int length);
   };
   
#endif
   
}

#endif
