/************************************************************************
* EPP RTK C++
* Copyright (C) 2001 Global Name Registry 
* 
* This library is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published
* by the Free Software Foundation; either version 2.1 of the License, or
* (at your option) any later version.
*
* This library is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Contact information:  epprtk@gnr.com
*
*                       EPP RTK
*                       GNR Ltd.
*                       125 High Holborn
*                       London WC1V 6QA
*                       United Kingdom
************************************************************************/

#ifndef GNR_REFCOUNT_H
#define GNR_REFCOUNT_H

//lint -emacro(717,REFCOUNT_DEBUG)
//#define REFCOUNT_DEBUG(__msg__) do { cerr << __FILE__ << " at line " << __LINE__ << ": " << __msg__  << endl;  } while (0);
#define REFCOUNT_DEBUG(__msg__) do {} while (0)
#define REFCOUNT_USE_TYPEINFO

#ifdef REFCOUNT_USE_TYPEINFO
#include <typeinfo>
#endif

#include "exceptions.h"

typedef domtools::Exception refcnt_exception;

/// This is what will be thrown if an attempt to assign the wrong type occurs
#define REFCOUNT_EXCEPTION_BADCAST (refcnt_exception(__FILE__,__LINE__,"Assignment between incompatible types"))
/// Thrown if an attempt to dereference a NULL reference
#define REFCOUNT_EXCEPTION_NULL (refcnt_exception(__FILE__,__LINE__,"Dereferencing NULL pointer"))


/** Simple reference counting class
 * 
 * The class supports multiple references. In cases where a single reference is needed,
 * std::auto_ptr<> may be more suitable.
 * 
 * Keep in mind the following overhead for this class: Both a pointer to the actual object
 * and a pointer to the counter is kept. This reduces dereferencing overhead, but at the
 * cost of sizeof(int) bytes extra storage every place a reference replaces a pointer.
 * 
 * In addition, size reference counters are allocated separately, your memory allocator
 * may waste lots of space (typically 16 extra bytes using Linux on x86). Using a vector
 * or similarly might alleviate this, but at the cost of more complexity. This latter
 * overhead is <i>per object referenced</i>, not per reference to the object.
 * 
 * Warning: It is possible to assign refcnt_ptr<A> to refcnt_ptr<B>, for different A and
 * B. Behaviour is only defined where the A is a subclass of B.
 * 
 * @bugs Not threadsafe. If you have references to the same object in multiple threads,
 *   you need to ensure yourself that the reference count is not modified in any way in
 *   multiple threads at the same time.
 */


template <typename T>
class refcnt_ptr
{
 public:
   refcnt_ptr(const void * const ptr = 0) : myPtr(0), myCount(0)
     {
	if (ptr != 0) throw(REFCOUNT_EXCEPTION_BADCAST);
	REFCOUNT_DEBUG("Creating NULL reference " << this << "of type " << typeid(*this).name());
     }
   
   refcnt_ptr(const refcnt_ptr<T> & right) : myPtr(0), myCount(0)
     {
	*this = right;
     }

   ~refcnt_ptr();

   refcnt_ptr<T>& operator=(const refcnt_ptr<T> & right)
     {
	if (this == &right) return *this;

	// Get rid of the previous object
	unref();
	
	//lint -e1555
	// myPtr and myCount should be shared between all refcnt_ptr objects, so
	// copying just the pointers is the right behaviour
	myPtr = right.myPtr;
	myCount = right.myCount;
	if (myCount) ++(*myCount);
	return *this;
     }
   
   explicit refcnt_ptr(T* ptr);

   template <typename R>
     operator refcnt_ptr<R> (void) const
     {
	refcnt_ptr<R> tmp;
	tmp.assign(*this);
	return tmp;
     }

   template <typename R>
     void assign(const refcnt_ptr<R> & rhs) 
     {
	T * ptr = dynamic_cast<T *>(rhs.getPtr());
	if (!ptr) throw(REFCOUNT_EXCEPTION_BADCAST);

	if ((void *)this == (void *)&rhs) return;
	
	// check that the references dont already point to the same object
	if (myCount != rhs.getCount()) {
	   // They dont
	   unref(); // Make sure we get rid of the previous object
	   myCount = rhs.getCount();
	   if (myCount) ++(*myCount);
	}
	myPtr = ptr;
     }


   
   /// For NULL assignment only. Anything else will throw an refcnt_exception for bad cast.
   refcnt_ptr& operator=(int rhs); 
   
   /// For NULL assignment only. Anything else will throw an refcnt_exception for bad cast.
   refcnt_ptr& operator=(const void *);
   
   /// Dereference reference to get at members or methods.
   T* operator->() const;
   
   T& operator*() const;
   
   int * getCount(void) const
     {
	return myCount;
     }

   /// Return the pointer this reference points to. Use with care.
   T* getPtr(void) const;
   
   /// Assign a new reference
   void ref(T * right);
   
   /// Drop the reference, possibly destroying the object
   void unref(void);

   /// Check for NULL referencel
   bool isNull(void) const;

   /// Check for equality of reference <b>NOT</b> the contents of the object
   bool operator==(const T * const right) const;

   /// Check for equality of reference <b>NOT</b> the contents of the object
   bool operator==(const refcnt_ptr & right) const;

   /// Check for non-equality of reference <b>NOT</b> the contents of the object
   bool operator!=(const T * const right) const;

   /// Check for non-equality of reference <b>NOT</b> the contents of the object
   bool operator!=(const refcnt_ptr & right) const;
 private:
   T * myPtr;
   int * myCount;
};





template <typename T>
refcnt_ptr<T>::refcnt_ptr(T* ptr)
{
   REFCOUNT_DEBUG("Constructing " << this << " of type " << typeid(*this).name() << " with count 1 (pointing to " << ptr << ")");
   myPtr = ptr;
   myCount = new int(1);
};


template <typename T>
refcnt_ptr<T>::~refcnt_ptr()
{
   try {
      REFCOUNT_DEBUG("destroying " << this);
      unref();
      myPtr = 0;   // To shut lint up. Freed in unref()
      myCount = 0; // To shut lint up. Freed in unref()
   } catch(...) {
   }
};



template <typename T>
refcnt_ptr<T>& refcnt_ptr<T>::operator=(int rhs)
{
   if (rhs != 0) throw(REFCOUNT_EXCEPTION_BADCAST);
   REFCOUNT_DEBUG("operator=: NULL assignment to ref of type " << typeid(*this).name() << ", replacing ref with count" << (myCount ? *myCount : 0));
   unref();
   return *this;
}


template <typename T>
refcnt_ptr<T>& refcnt_ptr<T>::operator=(const void * const rhs)
{
   if (rhs != 0) throw(REFCOUNT_EXCEPTION_BADCAST);
   REFCOUNT_DEBUG("operator=: NULL assignment to ref of type " << typeid(*this).name() << ", replacing ref with count" << (myCount ? *myCount : 0));
   unref();
   return *this;
}

template <typename T>
bool refcnt_ptr<T>::isNull(void) const
{
   return myPtr == 0;
}


template <typename T>
bool refcnt_ptr<T>::operator==(const T * const  right) const
{
   return myPtr == right;
};

template <typename T>
bool refcnt_ptr<T>::operator==(const refcnt_ptr<T> & right) const
{
   return (myPtr == right.myPtr) &&(myCount == right.myCount);
};

template <typename T>
bool refcnt_ptr<T>::operator!=(const T * const  right) const
{
   return myPtr != right;
};

template <typename T>
bool refcnt_ptr<T>::operator!=(const refcnt_ptr<T> & right) const
{
   return (myPtr != right.myPtr) || (myCount != right.myCount);
};


template <typename T>
T* refcnt_ptr<T>::operator->() const
{
   if (!myPtr) throw(REFCOUNT_EXCEPTION_NULL);
   return myPtr;
};

template <typename T>
T& refcnt_ptr<T>::operator*() const
{
   if (!myPtr) throw(REFCOUNT_EXCEPTION_NULL);
   return *myPtr;
};

template<typename T>
T * refcnt_ptr<T>::getPtr(void) const
{
   return myPtr;
};

template <typename T>
void refcnt_ptr<T>::ref(T * right)
{
   REFCOUNT_DEBUG("ref() to " << right << "of type " << typeid(*this).name() << ", replacing ref with count " << (myCount ? *myCount : 0));
   unref();
   myPtr = right;
   myCount = new int(1);
};

template <typename T>
void refcnt_ptr<T>::unref(void) 
{
   REFCOUNT_DEBUG("unref() of type " << typeid(*this).name() << ", with count " << (myCount ? *myCount : 0));
   if (myCount) {
      --(*myCount);
      if (*myCount == 0) {
	 REFCOUNT_DEBUG("  deleting ptr " << myPtr);
	 delete myPtr;
	 delete myCount;
      }
      myCount = 0;
   }
   myPtr = 0;
};



#endif

